from __future__ import annotations

from typing import TYPE_CHECKING

from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from rich.console import Console

from cyberdrop_dl.utils.args.browser_cookie_extraction import get_forum_cookies
from cyberdrop_dl.managers.manager import Manager

if TYPE_CHECKING:
    from typing import Dict

    from cyberdrop_dl.managers.manager import Manager

console = Console()


def edit_authentication_values_prompt(manager: Manager) -> None:
    """Edit the authentication values"""
    auth = manager.config_manager.authentication_data

    while True:
        console.clear()
        console.print("Editing Authentication Values")
        action = inquirer.select(
            message="What would you like to do?",
            choices=[
                Choice(1, "Edit Forum Authentication Values"),
                Choice(2, "Edit JDownloader Authentication Values"),
                Choice(3, "Edit Reddit Authentication Values"),
                Choice(4, "Edit GoFile API Key"),
                Choice(5, "Edit Imgur Client ID"),
                Choice(6, "Edit PixelDrain API Key"),
                Choice(7, "Done"),
            ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
            vi_mode=manager.vi_mode,
        ).execute()

        # Edit Forums
        if action == 1:
            edit_forum_authentication_values_prompt(manager)

        # Edit JDownloader
        elif action == 2:
            edit_jdownloader_authentication_values_prompt(auth)

        # Edit Reddit
        elif action == 3:
            edit_reddit_authentication_values_prompt(auth)

        # Edit GoFile API Key
        elif action == 4:
            console.clear()
            gofile_api_key = inquirer.text(
                message="Enter the GoFile API Key:",
                default=auth["GoFile"]["gofile_api_key"],
                long_instruction="You can get your premium GoFile API Key from https://gofile.io/myProfile",
                vi_mode=manager.vi_mode,
            ).execute()
            auth["GoFile"]["gofile_api_key"] = gofile_api_key

        # Edit Imgur Client ID
        elif action == 5:
            console.clear()
            imgur_client_id = inquirer.text(
                message="Enter the Imgur Client ID:",
                default=auth["Imgur"]["imgur_client_id"],
                long_instruction="You can create an app and get your client ID "
                                 "from https://imgur.com/account/settings/apps",
                vi_mode=manager.vi_mode,
            ).execute()
            auth["Imgur"]["imgur_client_id"] = imgur_client_id

        # Edit PixelDrain API Key
        elif action == 6:
            console.clear()
            pixeldrain_api_key = inquirer.text(
                message="Enter the PixelDrain API Key:",
                default=auth["PixelDrain"]["pixeldrain_api_key"],
                long_instruction="You can get your premium API Key from https://pixeldrain.com/user/api_keys",
                vi_mode=manager.vi_mode,
            ).execute()
            auth["PixelDrain"]["pixeldrain_api_key"] = pixeldrain_api_key

        # Done
        elif action == 7:
            manager.config_manager.write_updated_authentication_config()
            return


def edit_forum_authentication_values_prompt(manager: Manager) -> None:
    """Edit the forum authentication values"""
    while True:
        console.clear()
        console.print("Editing Forum Authentication Values")
        action = inquirer.select(
            message="What would you like to do?",
            choices=[
                Choice(1, "Browser Cookie Extraction"),
                Choice(2, "Enter Cookie Values Manually"),
                Choice(3, "Done"),
            ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
            vi_mode=manager.vi_mode,
        ).execute()

        # Browser Cookie Extraction
        if action == 1:
            action = inquirer.select(
                message="Which browser should we load cookies from?",
                choices=[
                    Choice("chrome", "Chrome"),
                    Choice("firefox", "FireFox"),
                    Choice("edge", "Edge"),
                    Choice("safari", "Safari"),
                    Choice("opera", "Opera"),
                    Choice("brave", "Brave"),
                    Choice(1, "Done"),
                ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
                vi_mode=manager.vi_mode,
            ).execute()

            # Done
            if action == 1:
                continue

            # Browser Selection
            if action == "chrome":
                get_forum_cookies(manager, "chrome")
            elif action == "firefox":
                get_forum_cookies(manager, "firefox")
            elif action == "edge":
                get_forum_cookies(manager, "edge")
            elif action == "safari":
                get_forum_cookies(manager, "safari")
            elif action == "opera":
                get_forum_cookies(manager, "opera")
            elif action == "brave":
                get_forum_cookies(manager, "brave")
            return

        # Enter Cred Values Manually
        elif action == 2:
            celebforum_username = inquirer.text(
                message="Enter your CelebForum Username:",
                default=manager.config_manager.authentication_data["Forums"]["celebforum_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            celebforum_password = inquirer.text(
                message="Enter your CelebForum Password:",
                default=manager.config_manager.authentication_data["Forums"]["celebforum_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            f95zone_username = inquirer.text(
                message="Enter your F95Zone Username:",
                default=manager.config_manager.authentication_data["Forums"]["f95zone_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            f95zone_password = inquirer.text(
                message="Enter your F95Zone Password:",
                default=manager.config_manager.authentication_data["Forums"]["f95zone_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            leakedmodels_username = inquirer.text(
                message="Enter your LeakedModels Username:",
                default=manager.config_manager.authentication_data["Forums"]["leakedmodels_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            leakedmodels_password = inquirer.text(
                message="Enter your LeakedModels Password:",
                default=manager.config_manager.authentication_data["Forums"]["leakedmodels_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            nudostar_username = inquirer.text(
                message="Enter your NudoStar Username:",
                default=manager.config_manager.authentication_data["Forums"]["nudostar_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            nudostar_password = inquirer.text(
                message="Enter your NudoStar Password:",
                default=manager.config_manager.authentication_data["Forums"]["nudostar_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            simpcity_username = inquirer.text(
                message="Enter your SimpCity Username:",
                default=manager.config_manager.authentication_data["Forums"]["simpcity_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            simpcity_password = inquirer.text(
                message="Enter your SimpCity Password:",
                default=manager.config_manager.authentication_data["Forums"]["simpcity_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            socialmediagirls_username = inquirer.text(
                message="Enter your SocialMediaGirls Username:",
                default=manager.config_manager.authentication_data["Forums"]["socialmediagirls_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            socialmediagirls_password = inquirer.text(
                message="Enter your SocialMediaGirls Password:",
                default=manager.config_manager.authentication_data["Forums"]["socialmediagirls_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            xbunker_username = inquirer.text(
                message="Enter your XBunker Username:",
                default=manager.config_manager.authentication_data["Forums"]["xbunker_username"],
                vi_mode=manager.vi_mode,
            ).execute()
            xbunker_password = inquirer.text(
                message="Enter your XBunker Password:",
                default=manager.config_manager.authentication_data["Forums"]["xbunker_password"],
                vi_mode=manager.vi_mode,
            ).execute()

            manager.config_manager.authentication_data["Forums"]["celebforum_username"] = celebforum_username
            manager.config_manager.authentication_data["Forums"]["f95zone_username"] = f95zone_username
            manager.config_manager.authentication_data["Forums"]["leakedmodels_username"] = leakedmodels_username
            manager.config_manager.authentication_data["Forums"]["nudostar_username"] = nudostar_username
            manager.config_manager.authentication_data["Forums"]["simpcity_username"] = simpcity_username
            manager.config_manager.authentication_data["Forums"]["socialmediagirls_username"] = socialmediagirls_username
            manager.config_manager.authentication_data["Forums"]["xbunker_username"] = xbunker_username

            manager.config_manager.authentication_data["Forums"]["celebforum_password"] = celebforum_password
            manager.config_manager.authentication_data["Forums"]["f95zone_password"] = f95zone_password
            manager.config_manager.authentication_data["Forums"]["leakedmodels_password"] = leakedmodels_password
            manager.config_manager.authentication_data["Forums"]["nudostar_password"] = nudostar_password
            manager.config_manager.authentication_data["Forums"]["simpcity_password"] = simpcity_password
            manager.config_manager.authentication_data["Forums"]["socialmediagirls_password"] = socialmediagirls_password
            manager.config_manager.authentication_data["Forums"]["xbunker_password"] = xbunker_password
            return
        elif action == 3:
            return


def edit_jdownloader_authentication_values_prompt(auth: Dict) -> None:
    """Edit the JDownloader authentication values"""
    console.clear()
    jdownloader_username = inquirer.text(
        message="Enter the JDownloader Username:",
        default=auth["JDownloader"]["jdownloader_username"],
    ).execute()
    jdownloader_password = inquirer.text(
        message="Enter the JDownloader Password:",
        default=auth["JDownloader"]["jdownloader_password"],
    ).execute()
    jdownloader_device = inquirer.text(
        message="Enter the JDownloader Device Name:",
        default=auth["JDownloader"]["jdownloader_device"],
    ).execute()

    auth["JDownloader"]["jdownloader_username"] = jdownloader_username
    auth["JDownloader"]["jdownloader_password"] = jdownloader_password
    auth["JDownloader"]["jdownloader_device"] = jdownloader_device


def edit_reddit_authentication_values_prompt(auth: Dict) -> None:
    """Edit the reddit authentication values"""
    console.clear()
    console.print(
        "You can create a Reddit App to use here: https://www.reddit.com/prefs/apps/"
    )
    reddit_secret = inquirer.text(
        message="Enter the Reddit Secret value:",
        default=auth["Reddit"]["reddit_secret"],
    ).execute()
    reddit_personal_use_script = inquirer.text(
        message="Enter the Reddit Personal Use Script value:",
        default=auth["Reddit"]["reddit_personal_use_script"],
    ).execute()

    auth["Reddit"]["reddit_secret"] = reddit_secret
    auth["Reddit"]["reddit_personal_use_script"] = reddit_personal_use_script



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.separator import Separator
from InquirerPy.validator import EmptyInputValidator, PathValidator
from rich.console import Console

from cyberdrop_dl.utils.transfer.transfer_v4_config import transfer_v4_config
from cyberdrop_dl.utils.transfer.transfer_v4_db import transfer_v4_db

if TYPE_CHECKING:
    from typing import List

    from cyberdrop_dl.managers.manager import Manager


console = Console()


def main_prompt(manager: Manager) -> int:
    """Main prompt for the program"""
    action = inquirer.select(
        message="What would you like to do?",
        choices=[
            Choice(1, "Download"),
            Choice(2, "Download (All Configs)"),
            Choice(3, "Retry Failed Downloads"),
            Choice(4, "Sort All Configs"),
            Choice(5, "Edit URLs File"),
            Separator(),
            Choice(6, f"Select Config (Current: {manager.config_manager.loaded_config})"),
            Choice(7, "Change URLs.txt file and Download Location"),
            Choice(8, "Manage Configs"),
            Separator(),
            Choice(9, "Import Cyberdrop_V4 Items"),
            Choice(10, "Donate"),
            Choice(11, "Exit"),
        ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
        vi_mode=manager.vi_mode,
    ).execute()

    return action


def manage_configs_prompt(manager: Manager) -> int:
    """Manage Configs Prompt"""
    console.clear()
    action = inquirer.select(
        message="What would you like to do?",
        choices=[
            Choice(1, "Change Default Config"),
            Choice(2, "Create A New Config"),
            Choice(3, "Delete A Config"),
            Separator(),
            Choice(4, "Edit Config"),
            Choice(5, "Edit Authentication Values"),
            Choice(6, "Edit Global Values"),
            Choice(7, "Done"),
        ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
        vi_mode=manager.vi_mode,
    ).execute()

    return action


def select_config_prompt(manager: Manager, configs: List) -> str:
    """Select a config file from a list of configs"""
    choice = inquirer.fuzzy(
        choices=configs,
        multiselect=False,
        validate=lambda result: len(result) > 0,
        invalid_message="Need to select a config.",
        message="Select a config file:",
        long_instruction="ARROW KEYS: Navigate | TYPE: Filter | TAB: select, ENTER: Finish Selection",
        vi_mode=manager.vi_mode,
    ).execute()

    return choice


def import_cyberdrop_v4_items_prompt(manager: Manager) -> None:
    """Import Cyberdrop_V4 Items"""
    while True:
        console.clear()
        console.print("Editing Config Values")
        action = inquirer.select(
            message="What would you like to do?",
            choices=[
                Choice(1, "Import Config"),
                Choice(2, "Import download_history.sql"),
                Choice(3, "Done"),
            ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
            vi_mode=manager.vi_mode,
        ).execute()

        # Import Config
        if action == 1:
            new_config_name = inquirer.text(
                message="What should this config be called?",
                validate=EmptyInputValidator("Input should not be empty"),
                vi_mode=manager.vi_mode,
            ).execute()

            if (manager.path_manager.config_dir / new_config_name).is_dir():
                console.print(f"Config with name '{new_config_name}' already exists!")
                inquirer.confirm(message="Press enter to return to the import menu.").execute()
                continue

            home_path = "~/" if os.name == "posix" else "C:\\"
            import_config_path = inquirer.filepath(
                message="Select the config file to import",
                default=home_path,
                validate=PathValidator(is_file=True, message="Input is not a file"),
            ).execute()

            transfer_v4_config(manager, import_config_path, new_config_name)

        # Import download_history.sql
        elif action == 2:
            home_path = "~/" if os.name == "posix" else "C:\\"
            import_download_history_path = inquirer.filepath(
                message="Select the download_history.sql file to import",
                default=home_path,
                validate=PathValidator(is_file=True, message="Input is not a file"),
                vi_mode=manager.vi_mode,
            ).execute()

            transfer_v4_db(import_download_history_path, manager.path_manager.history_db)

        # Done
        elif action == 3:
            break


def donations_prompt(manager: Manager) -> None:
    """Donations prompt"""
    console.clear()
    console.print("[bold]Donations[/bold]")
    console.print("")
    console.print("I started making this program around three years ago at this point,"
                  "\nIt has grown larger than I could've imagined and I'm very proud of it."
                  "\nI have put a lot of time and effort into this program and I'm glad that people are using it."
                  "\nThanks to everyone that have supported me, "
                  "it keeps me motivated to continue working on this program.")
    console.print("")
    console.print("If you'd like to support me and my work, you can donate to me via the following methods:")
    console.print("BuyMeACoffee: https://www.buymeacoffee.com/juleswinnft")
    console.print("Github Sponsor: https://github.com/sponsors/Jules-WinnfieldX")

    console.print("")
    console.print("Thank you for your support!")
    console.print("")
    inquirer.confirm(message="Press enter to return to the main menu.", vi_mode=manager.vi_mode).execute()



from typing import TYPE_CHECKING

import aiofiles

if TYPE_CHECKING:
    from pathlib import Path
    from yarl import URL

    from cyberdrop_dl.managers.manager import Manager


class LogManager:
    def __init__(self, manager: 'Manager'):
        self.manager = manager
        self.main_log: Path = manager.path_manager.main_log
        self.last_post_log: Path = manager.path_manager.last_post_log
        self.unsupported_urls_log: Path = manager.path_manager.unsupported_urls_log
        self.download_error_log: Path = manager.path_manager.download_error_log
        self.scrape_error_log: Path = manager.path_manager.scrape_error_log

    def startup(self) -> None:
        """Startup process for the file manager"""
        self.main_log.unlink(missing_ok=True)
        self.main_log.touch(exist_ok=True)
        self.last_post_log.unlink(missing_ok=True)
        self.last_post_log.touch(exist_ok=True)
        self.unsupported_urls_log.unlink(missing_ok=True)
        self.unsupported_urls_log.touch(exist_ok=True)
        self.download_error_log.unlink(missing_ok=True)
        self.download_error_log.touch(exist_ok=True)
        self.scrape_error_log.unlink(missing_ok=True)
        self.scrape_error_log.touch(exist_ok=True)

    async def write_last_post_log(self, url: 'URL') -> None:
        """Writes to the last post log"""
        async with aiofiles.open(self.last_post_log, 'a') as f:
            await f.write(f"{url}\n")

    async def write_unsupported_urls_log(self, url: 'URL') -> None:
        """Writes to the unsupported urls log"""
        async with aiofiles.open(self.unsupported_urls_log, 'a') as f:
            await f.write(f"{url}\n")

    async def write_download_error_log(self, url: 'URL', error_message: str) -> None:
        """Writes to the download error log"""
        async with aiofiles.open(self.download_error_log, 'a') as f:
            await f.write(f"{url},{error_message}\n")

    async def write_scrape_error_log(self, url: 'URL', error_message: str) -> None:
        """Writes to the scrape error log"""
        async with aiofiles.open(self.scrape_error_log, 'a') as f:
            await f.write(f"{url},{error_message}\n")
            
    async def update_last_forum_post(self) -> None:
        """Updates the last forum post"""
        input_file = self.manager.path_manager.input_file
        base_urls = []
        
        async with aiofiles.open(input_file, 'r') as f:
            current_urls = await f.readlines()
            
        for url in current_urls:
            if "http" not in url:
                continue
            if "post-" in url:
                url = url.rsplit("/", 1)[0]
            if not url.endswith("\n"):
                url += "\n"
            base_urls.append(url)

        last_post_file = self.last_post_log
        if not last_post_file.exists():
            return

        async with aiofiles.open(last_post_file, 'r') as f:
            new_urls = await f.readlines()

        for url in new_urls:
            url_temp = url.rsplit("/", 1)[0] + "\n"
            if url_temp in base_urls:
                base_urls.remove(url_temp)
                base_urls.append(url)

        async with aiofiles.open(input_file, 'w') as f:
            await f.writelines(base_urls)



from __future__ import annotations

import re
from dataclasses import Field
from pathlib import Path
from typing import TYPE_CHECKING

import aiofiles
from yarl import URL

from cyberdrop_dl.clients.errors import NoExtensionFailure, JDownloaderFailure
from cyberdrop_dl.downloader.downloader import Downloader
from cyberdrop_dl.scraper.jdownloader import JDownloader
from cyberdrop_dl.utils.dataclasses.url_objects import ScrapeItem, MediaItem
from cyberdrop_dl.utils.utilities import log, get_filename_and_ext, get_download_path

if TYPE_CHECKING:
    from typing import List

    from cyberdrop_dl.managers.manager import Manager


class ScrapeMapper:
    """This class maps links to their respective handlers, or JDownloader if they are unsupported"""
    def __init__(self, manager: Manager):
        self.manager = manager
        self.mapping = {"bunkrr": self.bunkrr, "celebforum": self.celebforum, "coomer": self.coomer,
                        "cyberdrop": self.cyberdrop, "cyberfile": self.cyberfile, "e-hentai": self.ehentai,
                        "erome": self.erome, "fapello": self.fapello, "f95zone": self.f95zone, "gofile": self.gofile,
                        "hotpic": self.hotpic, "ibb.co": self.imgbb, "imageban": self.imageban, "imgbox": self.imgbox,
                        "imgur": self.imgur, "img.kiwi": self.imgwiki, "jpg.church": self.jpgchurch,
                        "jpg.homes": self.jpgchurch, "jpg.fish": self.jpgchurch, "jpg.fishing": self.jpgchurch,
                        "jpg.pet": self.jpgchurch, "jpeg.pet": self.jpgchurch, "jpg1.su": self.jpgchurch,
                        "jpg2.su": self.jpgchurch, "jpg3.su": self.jpgchurch, "jpg4.su": self.jpgchurch,
                        "host.church": self.jpgchurch, "kemono": self.kemono, "leakedmodels": self.leakedmodels,
                        "mediafire": self.mediafire, "nudostar.com": self.nudostar, "nudostar.tv": self.nudostartv,
                        "omegascans": self.omegascans, "pimpandhost": self.pimpandhost, "pixeldrain": self.pixeldrain,
                        "postimg": self.postimg, "realbooru": self.realbooru, "reddit": self.reddit, 
                        "redd.it": self.reddit, "redgifs": self.redgifs, "rule34.xxx": self.rule34xxx, 
                        "rule34.xyz": self.rule34xyz, "saint": self.saint, "scrolller": self.scrolller, 
                        "simpcity": self.simpcity, "socialmediagirls": self.socialmediagirls, "toonily": self.toonily, 
                        "xbunker": self.xbunker, "xbunkr": self.xbunkr, "bunkr": self.bunkrr}
        self.existing_crawlers = {}
        self.no_crawler_downloader = Downloader(self.manager, "no_crawler")
        self.jdownloader = JDownloader(self.manager)

    async def bunkrr(self) -> None:
        """Creates a Bunkr Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.bunkrr_crawler import BunkrrCrawler
        self.existing_crawlers['bunkrr'] = BunkrrCrawler(self.manager)
        self.existing_crawlers['bunkr'] = self.existing_crawlers['bunkrr']

    async def celebforum(self) -> None:
        """Creates a CelebForum Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.celebforum_crawler import CelebForumCrawler
        self.existing_crawlers['celebforum'] = CelebForumCrawler(self.manager)

    async def coomer(self) -> None:
        """Creates a Coomer Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.coomer_crawler import CoomerCrawler
        self.existing_crawlers['coomer'] = CoomerCrawler(self.manager)

    async def cyberdrop(self) -> None:
        """Creates a Cyberdrop Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.cyberdrop_crawler import CyberdropCrawler
        self.existing_crawlers['cyberdrop'] = CyberdropCrawler(self.manager)

    async def cyberfile(self) -> None:
        """Creates a Cyberfile Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.cyberfile_crawler import CyberfileCrawler
        self.existing_crawlers['cyberfile'] = CyberfileCrawler(self.manager)

    async def ehentai(self) -> None:
        """Creates a EHentai Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.ehentai_crawler import EHentaiCrawler
        self.existing_crawlers['e-hentai'] = EHentaiCrawler(self.manager)

    async def erome(self) -> None:
        """Creates a Erome Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.erome_crawler import EromeCrawler
        self.existing_crawlers['erome'] = EromeCrawler(self.manager)

    async def fapello(self) -> None:
        """Creates a Fappelo Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.fapello_crawler import FapelloCrawler
        self.existing_crawlers['fapello'] = FapelloCrawler(self.manager)

    async def f95zone(self) -> None:
        """Creates a F95Zone Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.f95zone_crawler import F95ZoneCrawler
        self.existing_crawlers['f95zone'] = F95ZoneCrawler(self.manager)

    async def gofile(self) -> None:
        """Creates a GoFile Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.gofile_crawler import GoFileCrawler
        self.existing_crawlers['gofile'] = GoFileCrawler(self.manager)

    async def hotpic(self) -> None:
        """Creates a HotPic Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.hotpic_crawler import HotPicCrawler
        self.existing_crawlers['hotpic'] = HotPicCrawler(self.manager)

    async def imageban(self) -> None:
        """Creates a ImageBan Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.imageban_crawler import ImageBanCrawler
        self.existing_crawlers['imageban'] = ImageBanCrawler(self.manager)

    async def imgbb(self) -> None:
        """Creates a ImgBB Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.imgbb_crawler import ImgBBCrawler
        self.existing_crawlers['ibb.co'] = ImgBBCrawler(self.manager)

    async def imgbox(self) -> None:
        """Creates a ImgBox Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.imgbox_crawler import ImgBoxCrawler
        self.existing_crawlers['imgbox'] = ImgBoxCrawler(self.manager)

    async def imgur(self) -> None:
        """Creates a Imgur Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.imgur_crawler import ImgurCrawler
        self.existing_crawlers['imgur'] = ImgurCrawler(self.manager)

    async def imgwiki(self) -> None:
        """Creates a ImgWiki Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.imgkiwi_crawler import ImgKiwiCrawler
        self.existing_crawlers['img.kiwi'] = ImgKiwiCrawler(self.manager)

    async def jpgchurch(self) -> None:
        """Creates a JPGChurch Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.jpgchurch_crawler import JPGChurchCrawler
        self.existing_crawlers['jpg.church'] = JPGChurchCrawler(self.manager)
        self.existing_crawlers['jpg.homes'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg.fish'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg.fishing'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg.pet'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpeg.pet'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg1.su'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg2.su'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg3.su'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['jpg4.su'] = self.existing_crawlers['jpg.church']
        self.existing_crawlers['host.church'] = self.existing_crawlers['jpg.church']

    async def kemono(self) -> None:
        """Creates a Kemono Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.kemono_crawler import KemonoCrawler
        self.existing_crawlers['kemono'] = KemonoCrawler(self.manager)

    async def leakedmodels(self) -> None:
        """Creates a LeakedModels Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.leakedmodels_crawler import LeakedModelsCrawler
        self.existing_crawlers['leakedmodels'] = LeakedModelsCrawler(self.manager)

    async def mediafire(self) -> None:
        """Creates a MediaFire Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.mediafire_crawler import MediaFireCrawler
        self.existing_crawlers['mediafire'] = MediaFireCrawler(self.manager)

    async def nudostar(self) -> None:
        """Creates a NudoStar Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.nudostar_crawler import NudoStarCrawler
        self.existing_crawlers['nudostar.com'] = NudoStarCrawler(self.manager)

    async def nudostartv(self) -> None:
        """Creates a NudoStarTV Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.nudostartv_crawler import NudoStarTVCrawler
        self.existing_crawlers['nudostar.tv'] = NudoStarTVCrawler(self.manager)

    async def omegascans(self) -> None:
        """Creates a OmegaScans Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.omegascans_crawler import OmegaScansCrawler
        self.existing_crawlers['omegascans'] = OmegaScansCrawler(self.manager)

    async def pimpandhost(self) -> None:
        """Creates a PimpAndHost Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.pimpandhost_crawler import PimpAndHostCrawler
        self.existing_crawlers['pimpandhost'] = PimpAndHostCrawler(self.manager)

    async def pixeldrain(self) -> None:
        """Creates a PixelDrain Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.pixeldrain_crawler import PixelDrainCrawler
        self.existing_crawlers['pixeldrain'] = PixelDrainCrawler(self.manager)

    async def postimg(self) -> None:
        """Creates a PostImg Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.postimg_crawler import PostImgCrawler
        self.existing_crawlers['postimg'] = PostImgCrawler(self.manager)
        
    async def realbooru(self) -> None:
        """Creates a RealBooru Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.realbooru_crawler import RealBooruCrawler
        self.existing_crawlers['realbooru'] = RealBooruCrawler(self.manager)

    async def reddit(self) -> None:
        """Creates a Reddit Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.reddit_crawler import RedditCrawler
        self.existing_crawlers['reddit'] = RedditCrawler(self.manager)
        self.existing_crawlers['redd.it'] = self.existing_crawlers['reddit']

    async def redgifs(self) -> None:
        """Creates a RedGifs Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.redgifs_crawler import RedGifsCrawler
        self.existing_crawlers['redgifs'] = RedGifsCrawler(self.manager)

    async def rule34xxx(self) -> None:
        """Creates a Rule34XXX Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.rule34xxx_crawler import Rule34XXXCrawler
        self.existing_crawlers['rule34.xxx'] = Rule34XXXCrawler(self.manager)

    async def rule34xyz(self) -> None:
        """Creates a Rule34XYZ Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.rule34xyz_crawler import Rule34XYZCrawler
        self.existing_crawlers['rule34.xyz'] = Rule34XYZCrawler(self.manager)

    async def saint(self) -> None:
        """Creates a Saint Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.saint_crawler import SaintCrawler
        self.existing_crawlers['saint'] = SaintCrawler(self.manager)

    async def scrolller(self) -> None:
        """Creates a Scrolller Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.scrolller_crawler import ScrolllerCrawler
        self.existing_crawlers['scrolller'] = ScrolllerCrawler(self.manager)

    async def simpcity(self) -> None:
        """Creates a SimpCity Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.simpcity_crawler import SimpCityCrawler
        self.existing_crawlers['simpcity'] = SimpCityCrawler(self.manager)

    async def socialmediagirls(self) -> None:
        """Creates a SocialMediaGirls Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.socialmediagirls_crawler import SocialMediaGirlsCrawler
        self.existing_crawlers['socialmediagirls'] = SocialMediaGirlsCrawler(self.manager)

    async def toonily(self) -> None:
        """Creates a Toonily Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.toonily_crawler import ToonilyCrawler
        self.existing_crawlers['toonily'] = ToonilyCrawler(self.manager)

    async def xbunker(self) -> None:
        """Creates a XBunker Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.xbunker_crawler import XBunkerCrawler
        self.existing_crawlers['xbunker'] = XBunkerCrawler(self.manager)

    async def xbunkr(self) -> None:
        """Creates a XBunkr Crawler instance"""
        from cyberdrop_dl.scraper.crawlers.xbunkr_crawler import XBunkrCrawler
        self.existing_crawlers['xbunkr'] = XBunkrCrawler(self.manager)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def start_scrapers(self) -> None:
        """Starts all scrapers"""
        for key in self.mapping:
            await self.mapping[key]()
            crawler = self.existing_crawlers[key]
            await crawler.startup()

    async def start_jdownloader(self) -> None:
        """Starts JDownloader"""
        if self.jdownloader.enabled and isinstance(self.jdownloader.jdownloader_agent, Field):
            await self.jdownloader.jdownloader_setup()

    async def start(self) -> None:
        """Starts the orchestra"""
        self.manager.scrape_mapper = self

        await self.start_scrapers()
        await self.start_jdownloader()

        await self.no_crawler_downloader.startup()

        if not self.manager.args_manager.retry:
            await self.load_links()
        else:
            await self.load_failed_links()

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def regex_links(self, line: str) -> List:
        """Regex grab the links from the URLs.txt file
        This allows code blocks or full paragraphs to be copy and pasted into the URLs.txt"""
        yarl_links = []
        if line.lstrip().rstrip().startswith('#'):
            return yarl_links

        all_links = [x.group().replace(".md.", ".") for x in re.finditer(
            r"(?:http.*?)(?=($|\n|\r\n|\r|\s|\"|\[/URL]|']\[|]\[|\[/img]))", line)]
        for link in all_links:
            yarl_links.append(URL(link))
        return yarl_links

    async def load_links(self) -> None:
        """Loads links from args / input file"""
        input_file = self.manager.path_manager.input_file

        links = []
        if not self.manager.args_manager.other_links:
            block_quote = False
            async with aiofiles.open(input_file, "r", encoding="utf8") as f:
                async for line in f:
                    assert isinstance(line, str)
                    block_quote = not block_quote if line == "#\n" else block_quote
                    if not block_quote:
                        links.extend(await self.regex_links(line))
        else:
            links.extend(self.manager.args_manager.other_links)
        links = list(filter(None, links))

        if not links:
            await log("No valid links found.", 30)
        for link in links:
            item = ScrapeItem(url=link, parent_title="")
            self.manager.task_group.create_task(self.map_url(item))

    async def load_failed_links(self) -> None:
        """Loads failed links from db"""
        items = await self.manager.db_manager.history_table.get_failed_items()
        for item in items:
            link = URL(item[0])
            retry_path = Path(item[1])

            item = ScrapeItem(link, parent_title="", part_of_album=True, retry=True, retry_path=retry_path)
            self.manager.task_group.create_task(self.map_url(item))

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def extension_check(self, url: URL) -> bool:
        """Checks if the URL has a valid extension"""
        try:
            filename, ext = await get_filename_and_ext(url.name)

            from cyberdrop_dl.utils.utilities import FILE_FORMATS
            if ext in FILE_FORMATS['Images'] or ext in FILE_FORMATS['Videos'] or ext in FILE_FORMATS['Audio']:
                return True
            return False
        except NoExtensionFailure:
            return False

    async def map_url(self, scrape_item: ScrapeItem) -> None:
        """Maps URLs to their respective handlers"""
        if not scrape_item.url:
            return
        if not isinstance(scrape_item.url, URL):
            try:
                scrape_item.url = URL(scrape_item.url)
            except AttributeError:
                return
        try:
            if not scrape_item.url.host:
                return
        except AttributeError:
            return

        # blocked domains
        if any(x in scrape_item.url.host.lower() for x in ["facebook", "instagram", "fbcdn"]):
            await log(f"Skipping {scrape_item.url} as it is a blocked domain", 10)
            return

        skip = False
        if self.manager.config_manager.settings_data['Ignore_Options']['skip_hosts']:
            for skip_host in self.manager.config_manager.settings_data['Ignore_Options']['skip_hosts']:
                if skip_host in scrape_item.url.host:
                    skip = True
                    break
        if self.manager.config_manager.settings_data['Ignore_Options']['only_hosts']:
            skip = True
            for only_host in self.manager.config_manager.settings_data['Ignore_Options']['only_hosts']:
                if only_host in scrape_item.url.host:
                    skip = False
                    break

        if str(scrape_item.url).endswith("/"):
            if scrape_item.url.query_string:
                query = scrape_item.url.query_string[:-1]
                scrape_item.url = scrape_item.url.with_query(query)
            else:
                scrape_item.url = scrape_item.url.with_path(scrape_item.url.path[:-1])

        key = next((key for key in self.mapping if key in scrape_item.url.host.lower()), None)

        if key and not skip:
            scraper = self.existing_crawlers[key]
            self.manager.task_group.create_task(scraper.run(scrape_item))
            return

        elif skip:
            await log(f"Skipping URL by Config Selections: {scrape_item.url}", 10)

        elif await self.extension_check(scrape_item.url):
            check_complete = await self.manager.db_manager.history_table.check_complete("no_crawler", scrape_item.url, scrape_item.url)
            if check_complete:
                await log(f"Skipping {scrape_item.url} as it has already been downloaded", 10)
                await self.manager.progress_manager.download_progress.add_previously_completed()
                return
            await scrape_item.add_to_parent_title("Loose Files")
            scrape_item.part_of_album = True
            download_folder = await get_download_path(self.manager, scrape_item, "no_crawler")
            filename, ext = await get_filename_and_ext(scrape_item.url.name)
            media_item = MediaItem(scrape_item.url, scrape_item.url, None, download_folder, filename, ext, filename)
            self.manager.task_group.create_task(self.no_crawler_downloader.run(media_item))

        elif self.jdownloader.enabled:
            await log(f"Sending unsupported URL to JDownloader: {scrape_item.url}", 10)
            try:
                await self.jdownloader.direct_unsupported_to_jdownloader(scrape_item.url, scrape_item.parent_title)
            except JDownloaderFailure as e:
                await log(f"Failed to send {scrape_item.url} to JDownloader", 40)
                await log(e.message, 40)
                await self.manager.log_manager.write_unsupported_urls_log(scrape_item.url)

        else:
            await log(f"Unsupported URL: {scrape_item.url}", 30)
            await self.manager.log_manager.write_unsupported_urls_log(scrape_item.url)



from __future__ import annotations

import asyncio
import copy
import itertools
import os
from http import HTTPStatus
from functools import wraps, partial
from pathlib import Path
from typing import TYPE_CHECKING, Tuple

import aiofiles
import aiohttp
from aiohttp import ClientSession

from cyberdrop_dl.clients.errors import DownloadFailure, InvalidContentTypeFailure
from cyberdrop_dl.utils.utilities import FILE_FORMATS, log

if TYPE_CHECKING:
    from typing import Callable, Coroutine, Any

    from cyberdrop_dl.managers.client_manager import ClientManager
    from cyberdrop_dl.managers.manager import Manager
    from cyberdrop_dl.utils.dataclasses.url_objects import MediaItem


async def is_4xx_client_error(status_code: int) -> bool:
    """Checks whether the HTTP status code is 4xx client error"""
    if isinstance(status_code, str):
        return True
    return HTTPStatus.BAD_REQUEST <= status_code < HTTPStatus.INTERNAL_SERVER_ERROR


def limiter(func):
    """Wrapper handles limits for download session"""

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        domain_limiter = await self.client_manager.get_rate_limiter(args[0])
        await asyncio.sleep(await self.client_manager.get_downloader_spacer(args[0]))
        await self._global_limiter.acquire()
        await domain_limiter.acquire()

        async with aiohttp.ClientSession(headers=self._headers, raise_for_status=False,
                                         cookie_jar=self.client_manager.cookies, timeout=self._timeouts,
                                         trace_configs=self.trace_configs) as client:
            kwargs['client_session'] = client
            return await func(self, *args, **kwargs)
    return wrapper


class DownloadClient:
    """AIOHTTP operations for downloading"""
    def __init__(self, manager: Manager, client_manager: ClientManager):
        self.manager = manager
        self.client_manager = client_manager
        
        self._headers = {"user-agent": client_manager.user_agent}
        self._timeouts = aiohttp.ClientTimeout(total=client_manager.read_timeout + client_manager.connection_timeout,
                                               connect=client_manager.connection_timeout)
        self._global_limiter = self.client_manager.global_rate_limiter

        self.trace_configs = []
        if os.getenv("PYCHARM_HOSTED") is not None:
            async def on_request_start(session, trace_config_ctx, params):
                await log(f"Starting download {params.method} request to {params.url}", 40)

            async def on_request_end(session, trace_config_ctx, params):
                await log(f"Finishing download {params.method} request to {params.url}", 40)
                await log(f"Response status for {params.url}: {params.response.status}", 40)

            trace_config = aiohttp.TraceConfig()
            trace_config.on_request_start.append(on_request_start)
            trace_config.on_request_end.append(on_request_end)
            self.trace_configs.append(trace_config)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    @limiter
    async def _download(self, domain: str, manager: Manager, media_item: MediaItem,
                        save_content: Callable[[aiohttp.StreamReader], Coroutine[Any, Any, None]], client_session: ClientSession) -> None:
        """Downloads a file"""
        headers = copy.deepcopy(self._headers)
        headers['Referer'] = str(media_item.referer)
        if domain == "pixeldrain":
            if self.manager.config_manager.authentication_data['PixelDrain']['pixeldrain_api_key']:
                headers["Authorization"] = await self.manager.download_manager.basic_auth("Cyberdrop-DL", self.manager.config_manager.authentication_data['PixelDrain']['pixeldrain_api_key'])
        if isinstance(media_item.partial_file, Path):
            resume_point = media_item.partial_file.stat().st_size if media_item.partial_file.exists() else 0
            headers['Range'] = f'bytes={resume_point}-'

        await asyncio.sleep(self.client_manager.download_delay)

        async with client_session.get(media_item.url, headers=headers, ssl=self.client_manager.ssl_context,
                                      proxy=self.client_manager.proxy) as resp:
            if resp.status == HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE:
                media_item.partial_file.unlink()
                
            await self.client_manager.check_http_status(resp, download=True)
            content_type = resp.headers.get('Content-Type')
            
            if not isinstance(media_item.filesize, int):
                media_item.filesize = int(resp.headers.get('Content-Length', '0'))
            if not isinstance(media_item.partial_file, Path):
                proceed, skip = await self.get_final_file_info(media_item, domain)
                await self.mark_incomplete(media_item, domain)
                if not proceed:
                    await log(f"Skipping {media_item.url} as it has already been downloaded", 10)
                    await self.manager.progress_manager.download_progress.add_previously_completed(False)
                    await self.mark_completed(media_item, domain)
                    return
                if skip:
                    await self.manager.progress_manager.download_progress.add_skipped()
                    return
            
            ext = Path(media_item.filename).suffix.lower()
            if content_type and any(s in content_type.lower() for s in ('html', 'text')) and ext not in FILE_FORMATS['Text']:
                raise InvalidContentTypeFailure(message=f"Received {content_type}, was expecting other")

            if resp.status != HTTPStatus.PARTIAL_CONTENT and media_item.partial_file.is_file():
                media_item.partial_file.unlink()
                
            media_item.task_id = await self.manager.progress_manager.file_progress.add_task(f"({domain.upper()}) {media_item.filename}", media_item.filesize)
            if media_item.partial_file.is_file():
                resume_point = media_item.partial_file.stat().st_size
                await self.manager.progress_manager.file_progress.advance_file(media_item.task_id, resume_point)

            await save_content(resp.content)

    async def _append_content(self, media_item, content: aiohttp.StreamReader, update_progress: partial) -> None:
        """Appends content to a file"""
        if not await self.client_manager.manager.download_manager.check_free_space():
            raise DownloadFailure(status="No Free Space", message="Not enough free space")

        media_item.partial_file.parent.mkdir(parents=True, exist_ok=True)
        if not media_item.partial_file.is_file():
            media_item.partial_file.touch()
        async with aiofiles.open(media_item.partial_file, mode='ab') as f:
            async for chunk, _ in content.iter_chunks():
                await asyncio.sleep(0)
                await f.write(chunk)
                await update_progress(len(chunk))
        if not content.total_bytes and not media_item.partial_file.stat().st_size:
            media_item.partial_file.unlink()
            raise DownloadFailure(status=HTTPStatus.NOT_FOUND, message="File is empty")

    async def download_file(self, manager: Manager, domain: str, media_item: MediaItem) -> None:
        """Starts a file"""
        async def save_content(content: aiohttp.StreamReader) -> None:
            await self._append_content(media_item, content, partial(manager.progress_manager.file_progress.advance_file, media_item.task_id))

        await self._download(domain, manager, media_item, save_content)
        media_item.partial_file.rename(media_item.complete_file)
        await self.mark_completed(media_item, domain)
        
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    
    async def mark_incomplete(self, media_item: MediaItem, domain: str) -> None:
        """Marks the media item as incomplete in the database"""
        await self.manager.db_manager.history_table.insert_incompleted(domain, media_item)

    async def mark_completed(self, media_item: MediaItem, domain: str) -> None:
        """Marks the media item as completed in the database"""
        await self.manager.db_manager.history_table.mark_complete(domain, media_item)
    
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""
    
    async def get_download_dir(self, media_item: MediaItem) -> Path:
        """Returns the download directory for the media item"""
        download_folder = media_item.download_folder
        if self.manager.args_manager.retry:
            return download_folder

        if self.manager.config_manager.settings_data['Download_Options']['block_download_sub_folders']:
            while download_folder.parent != self.manager.path_manager.download_dir:
                download_folder = download_folder.parent
            media_item.download_folder = download_folder
        return download_folder
    
    async def get_final_file_info(self, media_item: MediaItem, domain: str) -> tuple[bool, bool]:
        """Complicated checker for if a file already exists, and was already downloaded"""
        download_dir = await self.get_download_dir(media_item)
        media_item.complete_file = download_dir / media_item.filename
        media_item.partial_file = media_item.complete_file.with_suffix(media_item.complete_file.suffix + '.part')
        
        expected_size = media_item.filesize if isinstance(media_item.filesize, int) else None
        proceed = True
        skip = False
        while True:
            if expected_size:
                file_size_check = await self.check_filesize_limits(media_item)
                if not file_size_check:
                    await log(f"Download Skip {media_item.url} due to filesize restrictions", 10)
                    proceed = False
                    skip = True
                    return proceed, skip

            if not media_item.complete_file.exists() and not media_item.partial_file.exists():
                break

            if media_item.complete_file.exists() and media_item.complete_file.stat().st_size == media_item.filesize:
                proceed = False
                break

            downloaded_filename = await self.manager.db_manager.history_table.get_downloaded_filename(domain, media_item)
            if not downloaded_filename:
                media_item.complete_file, media_item.partial_file = await self.iterate_filename(media_item.complete_file, media_item)
                break

            if media_item.filename == downloaded_filename:
                if media_item.partial_file.exists():
                    if media_item.partial_file.stat().st_size >= media_item.filesize != 0:
                        media_item.partial_file.unlink()
                    if media_item.partial_file.stat().st_size == media_item.filesize:
                        if media_item.complete_file.exists():
                            new_complete_filename, new_partial_file = await self.iterate_filename(media_item.complete_file, media_item)
                            media_item.partial_file.rename(new_complete_filename)
                            proceed = False

                            media_item.complete_file = new_complete_filename
                            media_item.partial_file = new_partial_file
                        else:
                            proceed = False
                            media_item.partial_file.rename(media_item.complete_file)
                elif media_item.complete_file.exists():
                    if media_item.complete_file.stat().st_size == media_item.filesize:
                        proceed = False
                    else:
                        media_item.complete_file, media_item.partial_file = await self.iterate_filename(media_item.complete_file, media_item)
                break

            media_item.filename = downloaded_filename
        media_item.download_filename = media_item.complete_file.name
        return proceed, skip

    async def iterate_filename(self, complete_file: Path, media_item: MediaItem) -> Tuple[Path, Path]:
        """Iterates the filename until it is unique"""
        partial_file = None
        for iteration in itertools.count(1):
            filename = f"{complete_file.stem} ({iteration}){media_item.ext}"
            temp_complete_file = media_item.download_folder / filename
            if not temp_complete_file.exists() and not await self.manager.db_manager.history_table.check_filename_exists(filename):
                media_item.filename = filename
                complete_file = media_item.download_folder / media_item.filename
                partial_file = complete_file.with_suffix(complete_file.suffix + '.part')
                break
        return complete_file, partial_file
    
    async def check_filesize_limits(self, media: MediaItem) -> bool:
        """Checks if the file size is within the limits"""
        max_video_filesize = self.manager.config_manager.settings_data['File_Size_Limits']['maximum_video_size']
        min_video_filesize = self.manager.config_manager.settings_data['File_Size_Limits']['minimum_video_size']
        max_image_filesize = self.manager.config_manager.settings_data['File_Size_Limits']['maximum_image_size']
        min_image_filesize = self.manager.config_manager.settings_data['File_Size_Limits']['minimum_image_size']
        max_other_filesize = self.manager.config_manager.settings_data['File_Size_Limits']['maximum_other_size']
        min_other_filesize = self.manager.config_manager.settings_data['File_Size_Limits']['minimum_other_size']

        if media.ext in FILE_FORMATS['Images']:
            if max_image_filesize and min_image_filesize:
                if media.filesize < min_image_filesize or media.filesize > max_image_filesize:
                    return False
            if media.filesize < min_image_filesize:
                return False
            if max_image_filesize and media.filesize > max_image_filesize:
                return False
        elif media.ext in FILE_FORMATS['Videos']:
            if max_video_filesize and min_video_filesize:
                if media.filesize < min_video_filesize or media.filesize > max_video_filesize:
                    return False
            if media.filesize < min_video_filesize:
                return False
            if max_video_filesize and media.filesize > max_video_filesize:
                return False
        else:
            if max_other_filesize and min_other_filesize:
                if media.filesize < min_other_filesize or media.filesize > max_other_filesize:
                    return False
            if media.filesize < min_other_filesize:
                return False
            if max_other_filesize and media.filesize > max_other_filesize:
                return False
        return True



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.validator import PathValidator, EmptyInputValidator, NumberValidator
from rich.console import Console

from cyberdrop_dl.utils.dataclasses.supported_domains import SupportedDomains

if TYPE_CHECKING:
    from typing import Dict

    from cyberdrop_dl.managers.manager import Manager

console = Console()


def create_new_config_prompt(manager: Manager) -> None:
    """Create a new config file"""
    console.clear()
    console.print("Create a new config file")
    config_name = inquirer.text(
        message="Enter the name of the config:",
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()
    if (manager.path_manager.config_dir / config_name).is_dir():
        console.print(f"Config with name '{config_name}' already exists!")
        inquirer.confirm(message="Press enter to return to the main menu.").execute()
        return
    manager.config_manager.change_config(config_name)
    edit_config_values_prompt(manager)


def edit_config_values_prompt(manager: Manager) -> None:
    """Edit the config values"""
    config = manager.config_manager.settings_data

    while True:
        console.clear()
        console.print("Editing Config Values")
        action = inquirer.select(
            message="What would you like to do?",
            choices=[
                Choice(1, "Edit Download Options"),
                Choice(2, "Edit Input / Output File Paths"),
                Choice(3, "Edit Log File Naming / Path"),
                Choice(4, "Edit File Size Limits"),
                Choice(5, "Edit Ignore Options"),
                Choice(6, "Edit Runtime Options"),
                Choice(7, "Edit Sorting Options"),
                Choice(8, "Done"),
            ], long_instruction="ARROW KEYS: Navigate | ENTER: Select",
            vi_mode=manager.vi_mode,
        ).execute()

        # Edit Download Options
        if action == 1:
            edit_download_options_prompt(manager, config)

        # Edit Input / Output File Paths
        elif action == 2:
            edit_input_output_file_paths_prompt(manager, config)

        # Edit Log File Naming / Path
        elif action == 3:
            edit_log_file_naming_path_prompt(manager, config)

        # Edit File Size Limits
        elif action == 4:
            edit_file_size_limits_prompt(manager, config)

        # Edit Ignore Options
        elif action == 5:
            edit_ignore_options_prompt(manager, config)

        # Edit Runtime Options
        elif action == 6:
            edit_runtime_options_prompt(manager, config)

        # Edit Sorting Options
        elif action == 7:
            edit_sort_options_prompt(manager, config)

        # Done
        elif action == 8:
            manager.config_manager.settings_data = config
            manager.config_manager.write_updated_settings_config()
            return


def edit_download_options_prompt(manager: Manager, config: Dict) -> None:
    """Edit the download options"""
    console.clear()
    action = inquirer.checkbox(
        message="Select the download options you want to enable:",
        choices=[
            Choice(value="block_download_sub_folders",
                   name="Block Download Sub Folders",
                   enabled=config["Download_Options"]["block_download_sub_folders"]),
            Choice(value="disable_download_attempt_limit",
                   name="Disable Download Attempt Limit",
                   enabled=config["Download_Options"]["disable_download_attempt_limit"]),
            Choice(value="disable_file_timestamps",
                   name="Disable File Timestamps Editing",
                   enabled=config["Download_Options"]["disable_file_timestamps"]),
            Choice(value="include_album_id_in_folder_name",
                   name="Include Album ID In Folder Name",
                   enabled=config["Download_Options"]["include_album_id_in_folder_name"]),
            Choice(value="include_thread_id_in_folder_name",
                   name="Include Thread ID In Folder Name",
                   enabled=config["Download_Options"]["include_album_id_in_folder_name"]),
            Choice(value="remove_domains_from_folder_names",
                   name="Remove Domains From Folder Names",
                   enabled=config["Download_Options"]["remove_domains_from_folder_names"]),
            Choice(value="remove_generated_id_from_filenames",
                   name="Remove Generated ID From Filenames",
                   enabled=config["Download_Options"]["remove_generated_id_from_filenames"]),
            Choice(value="scrape_single_forum_post",
                   name="Scrape Single Forum Post",
                   enabled=config["Download_Options"]["scrape_single_forum_post"]),
            Choice(value="separate_posts",
                   name="Separate Posts Into Folders",
                   enabled=config["Download_Options"]["separate_posts"]),
            Choice(value="skip_download_mark_completed",
                   name="Skip Download and Mark it as Completed",
                   enabled=config["Download_Options"]["skip_download_mark_completed"]),
        ], long_instruction="ARROW KEYS: Navigate | TAB: Select | ENTER: Confirm",
        vi_mode=manager.vi_mode,
    ).execute()

    for key in config["Download_Options"]:
        config["Download_Options"][key] = False

    for key in action:
        config["Download_Options"][key] = True


def edit_input_output_file_paths_prompt(manager: Manager, config: Dict) -> None:
    """Edit the input / output file paths"""
    console.clear()
    console.print("Editing Input / Output File Paths")
    input_file = inquirer.filepath(
        message="Enter the input file path:",
        default=str(config['Files']['input_file']),
        validate=PathValidator(is_file=True, message="Input is not a file"),
        vi_mode=manager.vi_mode,
    ).execute()
    download_folder = inquirer.text(
        message="Enter the download folder path:",
        default=str(config['Files']['download_folder']),
        validate=PathValidator(is_dir=True, message="Input is not a directory"),
        vi_mode=manager.vi_mode,
    ).execute()

    config['Files']['input_file'] = Path(input_file)
    config['Files']['download_folder'] = Path(download_folder)


def edit_log_file_naming_path_prompt(manager: Manager, config: Dict) -> None:
    """Edit the log file naming / path"""
    console.clear()
    console.print("Editing Log File Naming / Path")
    log_folder = inquirer.filepath(
        message="Enter the log folder path:",
        default=str(config['Logs']['log_folder']),
        validate=PathValidator(is_dir=True, message="Input is not a directory"),
        vi_mode=manager.vi_mode,
    ).execute()
    main_log_filename = inquirer.text(
        message="Enter the main log file name:",
        default=config['Logs']['main_log_filename'],
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()
    last_forum_post_filename = inquirer.text(
        message="Enter the last forum post log file name:",
        default=config['Logs']['last_forum_post_filename'],
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()
    unsupported_urls_filename = inquirer.text(
        message="Enter the unsupported urls log file name:",
        default=config['Logs']['unsupported_urls_filename'],
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()
    download_error_urls_filename = inquirer.text(
        message="Enter the download error urls log file name:",
        default=config['Logs']['download_error_urls_filename'],
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()
    scrape_error_urls_filename = inquirer.text(
        message="Enter the scrape error urls log file name:",
        default=config['Logs']['scrape_error_urls_filename'],
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()

    config['Logs']['log_folder'] = Path(log_folder)
    config['Logs']['main_log_filename'] = main_log_filename
    config['Logs']['last_forum_post_filename'] = last_forum_post_filename
    config['Logs']['unsupported_urls_filename'] = unsupported_urls_filename
    config['Logs']['download_error_urls_filename'] = download_error_urls_filename
    config['Logs']['scrape_error_urls_filename'] = scrape_error_urls_filename


def edit_file_size_limits_prompt(manager: Manager, config: Dict) -> None:
    """Edit the file size limits"""
    console.clear()
    console.print("Editing File Size Limits")
    maximum_image_size = inquirer.number(
        message="Enter the maximum image size:",
        default=int(config['File_Size_Limits']['maximum_image_size']),
        validate=NumberValidator(),
        long_instruction="This value is in bytes (0 is no limit)",
        vi_mode=manager.vi_mode,
    ).execute()
    maximum_video_size = inquirer.number(
        message="Enter the maximum video size:",
        default=int(config['File_Size_Limits']['maximum_video_size']),
        validate=NumberValidator(),
        long_instruction="This value is in bytes (0 is no limit)",
        vi_mode=manager.vi_mode,
    ).execute()
    maximum_other_size = inquirer.number(
        message="Enter the maximum other file type size:",
        default=int(config['File_Size_Limits']['maximum_other_size']),
        validate=NumberValidator(),
        long_instruction="This value is in bytes (0 is no limit)",
        vi_mode=manager.vi_mode,
    ).execute()
    minimum_image_size = inquirer.number(
        message="Enter the minimum image size:",
        default=int(config['File_Size_Limits']['minimum_image_size']),
        validate=NumberValidator(),
        long_instruction="This value is in bytes (0 is no limit)",
        vi_mode=manager.vi_mode,
    ).execute()
    minimum_video_size = inquirer.number(
        message="Enter the minimum video size:",
        default=int(config['File_Size_Limits']['minimum_video_size']),
        validate=NumberValidator(),
        long_instruction="This value is in bytes (0 is no limit)",
        vi_mode=manager.vi_mode,
    ).execute()
    minimum_other_size = inquirer.number(
        message="Enter the minimum other file type size:",
        default=int(config['File_Size_Limits']['minimum_other_size']),
        validate=NumberValidator(),
        long_instruction="This value is in bytes (0 is no limit)",
        vi_mode=manager.vi_mode,
    ).execute()

    config['File_Size_Limits']['maximum_image_size'] = int(maximum_image_size)
    config['File_Size_Limits']['maximum_video_size'] = int(maximum_video_size)
    config['File_Size_Limits']['maximum_other_size'] = int(maximum_other_size)
    config['File_Size_Limits']['minimum_image_size'] = int(minimum_image_size)
    config['File_Size_Limits']['minimum_video_size'] = int(minimum_video_size)
    config['File_Size_Limits']['minimum_other_size'] = int(minimum_other_size)


def edit_ignore_options_prompt(manager: Manager, config: Dict) -> None:
    """Edit the ignore options"""
    console.clear()
    console.print("Editing Ignore Options")
    action = inquirer.checkbox(
        message="Select the ignore options you want to enable:",
        choices=[
            Choice(value="exclude_videos",
                   name="Don't download videos files",
                   enabled=config["Ignore_Options"]["exclude_videos"]),
            Choice(value="exclude_images",
                   name="Don't download images files",
                   enabled=config["Ignore_Options"]["exclude_images"]),
            Choice(value="exclude_audio",
                   name="Don't download audio files",
                   enabled=config["Ignore_Options"]["exclude_audio"]),
            Choice(value="exclude_other",
                   name="Don't download other files",
                   enabled=config["Ignore_Options"]["exclude_other"]),
            Choice(value="ignore_coomer_ads",
                   name="Ignore coomer ads when scraping",
                   enabled=config["Ignore_Options"]["ignore_coomer_ads"]),
        ], long_instruction="ARROW KEYS: Move | TAB: Select | ENTER: Confirm",
        vi_mode=manager.vi_mode,
    ).execute()

    for key in config["Ignore_Options"]:
        config["Ignore_Options"][key] = False

    for key in action:
        config["Ignore_Options"][key] = True

    skip_choices = list(SupportedDomains.supported_hosts)
    skip_choices.insert(0, "None")
    skip_hosts = inquirer.fuzzy(
        choices=skip_choices,
        multiselect=True,
        message="Select any sites you want to ignore while scraping:",
        long_instruction="ARROW KEYS: Move | TYPE: Filter | TAB: Select | ENTER: Confirm",
        vi_mode=manager.vi_mode,
    ).execute()

    skip_hosts = [host for host in skip_hosts if host in SupportedDomains.supported_hosts]
    config["Ignore_Options"]["skip_hosts"] = skip_hosts

    only_choices = list(SupportedDomains.supported_hosts)
    only_choices.insert(0, "None")
    only_hosts = inquirer.fuzzy(
        choices=only_choices,
        multiselect=True,
        message="Select only the sites you want to scrape from:",
        long_instruction="ARROW KEYS: Move | TYPE: Filter | TAB: Select | ENTER: Confirm",
        vi_mode=manager.vi_mode,
    ).execute()

    only_hosts = [host for host in only_hosts if host in SupportedDomains.supported_hosts]
    config["Ignore_Options"]["only_hosts"] = only_hosts


def edit_runtime_options_prompt(manager: Manager, config: Dict) -> None:
    """Edit the runtime options"""
    console.clear()
    console.print("Editing Runtime Options")
    action = inquirer.checkbox(
        message="Select the runtime options you want to enable:",
        choices=[
            Choice(value="ignore_history",
                   name="Ignore the history (previously downloaded files)",
                   enabled=config["Runtime_Options"]["ignore_history"]),
            Choice(value="skip_check_for_partial_files",
                   name="Skip checking for partial files in the download folder",
                   enabled=config["Runtime_Options"]["skip_check_for_partial_files"]),
            Choice(value="skip_check_for_empty_folders",
                   name="Skip checking for empty folders in the download folder",
                   enabled=config["Runtime_Options"]["skip_check_for_empty_folders"]),
            Choice(value="delete_partial_files",
                   name="Delete partial files in the download folder",
                   enabled=config["Runtime_Options"]["delete_partial_files"]),
            Choice(value="send_unsupported_to_jdownloader",
                   name="Send unsupported urls to JDownloader to download",
                   enabled=config["Runtime_Options"]["send_unsupported_to_jdownloader"]),
            Choice(value="update_last_forum_post",
                   name="Update the last forum post after scraping",
                   enabled=config["Runtime_Options"]["update_last_forum_post"]),
        ], long_instruction="ARROW KEYS: Move | TAB: Select | ENTER: Confirm",
        vi_mode=manager.vi_mode,
    ).execute()

    log_level = inquirer.number(
        message="Enter the log level:",
        default=int(config['Runtime_Options']['log_level']),
        validate=NumberValidator(),
        long_instruction="10 is the default (uses pythons logging numerical levels)",
        vi_mode=manager.vi_mode,
    ).execute()

    ui_refresh_rate = inquirer.number(
        message="Enter the desired UI refresh rate:",
        default=int(config['Runtime_Options']['ui_refresh_rate']),
        min_allowed=1,
        validate=NumberValidator(),
        long_instruction="10 is the default",
        vi_mode=manager.vi_mode,
    ).execute()

    for key in config["Runtime_Options"]:
        config["Runtime_Options"][key] = False

    for key in action:
        config["Runtime_Options"][key] = True

    config['Runtime_Options']['log_level'] = int(log_level)
    config['Runtime_Options']['ui_refresh_rate'] = int(ui_refresh_rate)


def edit_sort_options_prompt(manager: Manager, config: Dict) -> None:
    """Edit the sort options"""
    console.clear()
    console.print("Editing Sort Options")
    config["Sorting"]["sort_downloads"] = False
    sort_downloads = inquirer.confirm(message="Do you want Cyberdrop-DL to sort files for you?", vi_mode=manager.vi_mode).execute()
    if sort_downloads:
        config["Sorting"]["sort_downloads"] = True
        sort_folder = inquirer.filepath(
            message="Enter the folder you want to sort files into:",
            default=str(config['Sorting']['sort_folder']),
            validate=PathValidator(is_dir=True, message="Input is not a directory"),
            vi_mode=manager.vi_mode,
        ).execute()
        sort_incremementer_format = inquirer.text(
            message="Enter the sort incrementer format:",
            default=config['Sorting']['sort_incremementer_format'],
            validate=EmptyInputValidator("Input should not be empty"),
            vi_mode=manager.vi_mode,
        ).execute()
        sorted_audio = inquirer.text(
            message="Enter the format you want to sort audio files into:",
            default=config['Sorting']['sorted_audio'],
            validate=EmptyInputValidator("Input should not be empty"),
            vi_mode=manager.vi_mode,
        ).execute()
        sorted_video = inquirer.text(
            message="Enter the format you want to sort video files into:",
            default=config['Sorting']['sorted_video'],
            validate=EmptyInputValidator("Input should not be empty"),
            vi_mode=manager.vi_mode,
        ).execute()
        sorted_image = inquirer.text(
            message="Enter the format you want to sort image files into:",
            default=config['Sorting']['sorted_image'],
            validate=EmptyInputValidator("Input should not be empty"),
            vi_mode=manager.vi_mode,
        ).execute()
        sorted_other = inquirer.text(
            message="Enter the format you want to sort other files into:",
            default=config['Sorting']['sorted_other'],
            validate=EmptyInputValidator("Input should not be empty"),
            vi_mode=manager.vi_mode,
        ).execute()

        config['Sorting']['sort_folder'] = Path(sort_folder)
        config['Sorting']['sort_incremementer_format'] = sort_incremementer_format
        config['Sorting']['sorted_audio'] = sorted_audio
        config['Sorting']['sorted_video'] = sorted_video
        config['Sorting']['sorted_image'] = sorted_image
        config['Sorting']['sorted_other'] = sorted_other
